/*
=====================================================================

  LEX.C: lexical analyzer for HTML parser
  Verion 1.0
  By Shuming Tan
  Copyright(c) Abraxas Software Inc. (R), 1995, All rights reserved

=====================================================================
*/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "public.h"
#include "html.h"

#define LEX_DEBUG 1
#define UNKNOWN_TAG   -1

#define LONGEST_CHAR_ENTITY           7
#define NON_BREAK_SPACE               0200
#define UP_AE_DIPHTONG                0201
#define UP_A_ACUTE_ACCENT             0202
#define UP_A_CIRCUMFLEX_ACCENT        0203
#define UP_A_GRAVE_ACCENT             0204
#define UP_A_RING                     0205
#define UP_A_TILDE                    0206
#define UP_A_DIERESIS_OR_UMLAUT_MARK  0207
#define UP_C_CEDILLA                  0210
#define UP_ETH_ICELANDIC              0211
#define UP_E_ACUTE_ACCENT             0212
#define UP_E_CIRCUMFLEX_ACCENT        0213
#define UP_E_GRAVE_ACCENT             0214
#define UP_E_DIERESIS_OR_UMLAUT_MARK  0215
#define UP_I_ACUTE_ACCENT             0216
#define UP_I_CIRCUMFLEX_ACCENT        0217
#define UP_I_GRAVE_ACCENT             0220
#define UP_I_DIERESIS_OR_UMLAUT_MARK  0221
#define UP_N_TILDE                    0222
#define UP_O_ACUTE_ACCENT             0223
#define UP_O_CIRCUMFLEX_ACCENT        0224
#define UP_O_GRAVE_ACCENT             0225
#define UP_O_SLASH                    0226
#define UP_O_TILDE                    0227
#define UP_O_DIERESIS_OR_UMLAUT_MARK  0230
#define UP_THONG_ICELANDIC            0231
#define UP_U_ACUTE_ACCENT             0232
#define UP_U_CIRCUMFLEX_ACCENT        0233
#define UP_U_GRAVE_ACCENT             0234
#define UP_U_DIERESIS_OR_UMLAUT_MARK  0235
#define UP_Y_ACUTE_ACCENT             0236
#define LOW_A_ACUTE_ACCENT            0237
#define LOW_A_CIRCUMFLEX_ACCENT       0240
#define LOW_AE_DIPHTONG               0241
#define LOW_A_GRAVE_ACCENT            0242
#define LOW_A_RING                    0243
#define LOW_A_TILDE                   0244
#define LOW_A_DIERESIS_OR_UMLAUT_MARK 0245
#define LOW_C_CEDILLA                 0246
#define LOW_E_ACUTE_ACCENT            0247
#define LOW_E_CIRCUMFLEX_ACCENT       0250
#define LOW_E_GRAVE_ACCENT            0251
#define LOW_ETH_ICELANDIC             0252
#define LOW_E_DIERESIS_OR_UMLAUT_MARK 0253
#define LOW_I_ACUTE_ACCENT            0254
#define LOW_I_CIRCUEMFLEX_ACCENT      0255
#define LOW_I_GRAVE_ACCENT            0256
#define LOW_I_DIERESIS_OR_UMLAUT_MARK 0257
#define LOW_N_TILDE                   0260
#define LOW_O_ACUTE_ACCENT            0261
#define LOW_O_CIRCUMFLEX_ACCENT       0262
#define LOW_O_GRAVE_ACCENT            0263
#define LOW_O_SLASH                   0264
#define LOW_O_TILDE                   0265
#define LOW_O_DIERESIS_OR_UMLAUT_MARK 0266
#define LOW_S_SHARP_GERMAN            0267
#define LOW_THORN_ICELANDIC           0270
#define LOW_U_ACUTE_ACCENT            0271
#define LOW_U_CIRCUMFLEX_ACCENT       0272
#define LOW_U_GRAVE_ACCENT            0273
#define LOW_U_DIERESIS_OR_UMLAUT_MARK 0274
#define LOW_Y_ACUTE_ACCENT            0275
#define LOW_Y_DIERESIS_OR_UMLAUT_MARK 0276
#define NOT_FOUND                     0377
#define YYERRCODE                     256
#define LINEWIDTH                     80

static struct        
{
  char n[20];
  int v;
} tagList[]=
{
  "A",          A_BEGIN_TAG,
  "/A",         A_END_TAG,
  "ADDRESS",    ADDRESS_BEGIN_TAG,
  "/ADDRESS",   ADDRESS_END_TAG,
  "B",          B_BEGIN_TAG,
  "/B",         B_END_TAG,
  "BASE",       BASE_TAG,
  "BLOCKQUOTE", BLOCKQUOTE_BEGIN_TAG,
  "/BLOCKQUOTE",BLOCKQUOTE_END_TAG,
  "BODY",       BODY_BEGIN_TAG,
  "/BODY",      BODY_END_TAG,
  "BR",         BR_TAG,
  "CITE",       CITE_BEGIN_TAG,
  "/CITE",      CITE_END_TAG,
  "CODE",       CODE_BEGIN_TAG,
  "/CODE",      CODE_END_TAG,
  "DD",         DD_TAG,
  "DIR",        DIR_BEGIN_TAG,
  "/DIR",       DIR_END_TAG,
  "DL",         DL_BEGIN_TAG,
  "/DL",        DL_END_TAG,
  "DT",         DT_TAG,
  "EM",         EM_BEGIN_TAG,
  "/EM",        EM_END_TAG,
  "FORM",       FORM_BEGIN_TAG,
  "/FORM",      FORM_END_TAG,
  "H1",         H1_BEGIN_TAG,
  "/H1",        H1_END_TAG,
  "H2",         H2_BEGIN_TAG,
  "/H2",        H2_END_TAG,
  "H3",         H3_BEGIN_TAG,
  "/H3",        H3_END_TAG,
  "H4",         H4_BEGIN_TAG,
  "/H4",        H4_END_TAG,
  "H5",         H5_BEGIN_TAG,
  "/H5",        H5_END_TAG,
  "H6",         H6_BEGIN_TAG,
  "/H6",        H6_END_TAG,
  "HEAD",       HEAD_BEGIN_TAG,
  "/HEAD",      HEAD_END_TAG,
  "HR",         HR_TAG,
  "HTML",       HTML_BEGIN_TAG,
  "/HTML",      HTML_END_TAG,
  "I",          I_BEGIN_TAG,
  "/I",         I_END_TAG,
  "IMG",        IMG_TAG,
  "INPUT",      INPUT_TAG,
  "ISINDEX",    ISINDEX_TAG,
  "KBD",        KBD_BEGIN_TAG,
  "/KBD",       KBD_END_TAG,
  "LI",         LI_TAG,
  "LINK",       LINK_TAG,
  "MENU",       MENU_BEGIN_TAG,
  "/MENU",      MENU_END_TAG,
  "OL",         OL_BEGIN_TAG,
  "/OL",        OL_END_TAG,
  "OPTION",     OPTION_TAG,
  "P",          P_TAG,
  "PRE",        PRE_BEGIN_TAG,
  "/PRE",       PRE_END_TAG,
  "SAMP",       SAMP_BEGIN_TAG,
  "/SAMP",      SAMP_END_TAG,
  "SELECT",     SELECT_BEGIN_TAG,
  "/SELECT",    SELECT_END_TAG,
  "STRONG",     STRONG_BEGIN_TAG,
  "/STRONG",    STRONG_END_TAG,
  "TEXTAREA",   TEXTAREA_BEGIN_TAG,
  "/TEXTAREA",  TEXTAREA_END_TAG,
  "TITLE",      TITLE_BEGIN_TAG,
  "/TITLE",     TITLE_END_TAG,
  "TT",         TT_BEGIN_TAG,
  "/TT",        TT_END_TAG,
  "UL",         UL_BEGIN_TAG,
  "/UL",        UL_END_TAG,
  "VAR",        VAR_BEGIN_TAG,
  "/VAR",       VAR_END_TAG,
  "PLAIN_TEXT", PLAIN_TEXT,
  "$EOT$",      UNKNOWN_TAG
};

static struct 
{
  char n[10];
  unsigned char v;
} specialChar[]=
{
  "&amp", '&',
  "&lt",  '<',
  "&gt",  '>',
  "&quot",'\"',
  "&nbsp",NON_BREAK_SPACE,
  "&AElig",UP_AE_DIPHTONG,
  "&Acute",UP_A_ACUTE_ACCENT,
  "&Acirc",UP_A_CIRCUMFLEX_ACCENT,
  "&Agrave",UP_A_GRAVE_ACCENT,
  "&Aring",UP_A_RING,
  "&Atilde",UP_A_TILDE,
  "&Auml",UP_A_DIERESIS_OR_UMLAUT_MARK,
  "&Ccedil",UP_C_CEDILLA,
  "&ETH",UP_ETH_ICELANDIC,
  "&Eacute",UP_E_ACUTE_ACCENT,
  "&Ecirc",UP_E_CIRCUMFLEX_ACCENT,
  "&Egrave",UP_E_GRAVE_ACCENT,
  "&Euml",UP_E_DIERESIS_OR_UMLAUT_MARK,
  "&Iacute",UP_I_ACUTE_ACCENT,
  "&Icirc",UP_I_CIRCUMFLEX_ACCENT,
  "&Igrave",UP_I_GRAVE_ACCENT,
  "&Iuml",UP_I_DIERESIS_OR_UMLAUT_MARK,
  "&Ntilde",UP_N_TILDE,
  "&Oacute",UP_O_ACUTE_ACCENT,
  "&Ocirc",UP_O_CIRCUMFLEX_ACCENT,
  "&Ograve",UP_O_GRAVE_ACCENT,
  "&Oslash",UP_O_SLASH,
  "&Otilde",UP_O_TILDE,
  "&Ouml",UP_O_DIERESIS_OR_UMLAUT_MARK,
  "&THONG",UP_THONG_ICELANDIC,
  "&Uacute",UP_U_ACUTE_ACCENT,
  "&Ucirc",UP_U_CIRCUMFLEX_ACCENT,
  "&Ugrave",UP_U_GRAVE_ACCENT,
  "&Uuml",UP_U_DIERESIS_OR_UMLAUT_MARK,
  "&Yacute",UP_Y_ACUTE_ACCENT,
  "&aacute",LOW_A_ACUTE_ACCENT,
  "&acirc",LOW_A_CIRCUMFLEX_ACCENT,
  "&aelig",LOW_AE_DIPHTONG,
  "&agrave",LOW_A_GRAVE_ACCENT,
  "&aring",LOW_A_RING,
  "&atilde",LOW_A_TILDE,
  "&auml",LOW_A_DIERESIS_OR_UMLAUT_MARK,
  "&ccdeil",LOW_C_CEDILLA,
  "&eacute",LOW_E_ACUTE_ACCENT,
  "&ecirc",LOW_E_CIRCUMFLEX_ACCENT,
  "&egrave",LOW_E_GRAVE_ACCENT,
  "&eth",LOW_ETH_ICELANDIC,
  "&euml",LOW_E_DIERESIS_OR_UMLAUT_MARK,
  "&iacute",LOW_I_ACUTE_ACCENT,
  "&icirc",LOW_I_CIRCUEMFLEX_ACCENT,
  "&igrave",LOW_I_GRAVE_ACCENT,
  "&iuml",LOW_I_DIERESIS_OR_UMLAUT_MARK,
  "&ntilde",LOW_N_TILDE,
  "&oacute",LOW_O_ACUTE_ACCENT,
  "&ocirc",LOW_O_CIRCUMFLEX_ACCENT,
  "&ograve",LOW_O_GRAVE_ACCENT,
  "&oslash",LOW_O_SLASH,
  "&otilde",LOW_O_TILDE,
  "&ouml",LOW_O_DIERESIS_OR_UMLAUT_MARK,
  "&szlig",LOW_S_SHARP_GERMAN,
  "&thorn",LOW_THORN_ICELANDIC,
  "&uacute",LOW_U_ACUTE_ACCENT,
  "&ucirc",LOW_U_CIRCUMFLEX_ACCENT,
  "&ugrave",LOW_U_GRAVE_ACCENT,
  "&uuml",LOW_U_DIERESIS_OR_UMLAUT_MARK,
  "&yacute",LOW_Y_ACUTE_ACCENT,
  "&yuml",LOW_Y_DIERESIS_OR_UMLAUT_MARK,
  "$EOT$", NOT_FOUND
};

struct tag_param *s, *p, *q;
extern int c;
extern char yyerrsrc[];
extern int yylineno;
extern int yyerrcnt;
extern FILE *yyin;
FILE *yyerrfile=stderr;
char *yydisplay(int tokenValue);
static char yytext[2048];
static char parmtoken[1024];

/* 
   given a string, search through the table to see if this is
   an legitimate tag, if so, return the corresponding integer
   representing this tag, otherwise, report unknown tag.
*/
int searchTag(char *s)
{
  int i;
  
  i=0;
  while (strcmp(tagList[i].n,"$EOT$"))
  {
    if (!strcmp(s,tagList[i].n))
    {
      return (tagList[i].v);
    }
    i++;
  }
  return (UNKNOWN_TAG);
}

/* 
   simlar to the searchTag(), search a character entity from table
   and return corresponding character's value 
*/
int searchChar(char *s)
{
  int i;

  i=0;
  while (strcmp(specialChar[i].n,"$EOT$"))
  {
    if (!strcmp(s,specialChar[i].n))
    {
      return (specialChar[i].v);
    }
    i++;
  }
  return (NOT_FOUND);
}

/* 
   since the browser is case insensitive, we need to convert
   the collected string into uppercases.
*/
void captalize(char *s)
{
  int i;

  i=0;
  while (s[i]!=0)
  {
    if (islower(s[i]))
    {
      s[i]='A'+s[i]-'a';
    }
    i++;
  }
}

int yylex()
{
  int tokenValue;

  tokenValue=yylex1();
#if (LEX_DEBUG)
  printf("tokenValue=[%d],actual token=[%s]\n",tokenValue,yydisplay(tokenValue));
#endif
  return (tokenValue);
}

/* Real Lexical Analyzer Function */
int yylex1()
{
  int goback; /* marks if the lexer need to continue scanning */
  int isString;
  int c1,c2;
  int t;
  int i,j;
  int ch;
  char esc[LONGEST_CHAR_ENTITY];
  static int isPreformattedText=0;

  goback=0;
  while (1)
  {
    if (isspace(c)&&!isPreformattedText)
    {
      while (isspace(c))
      {
        if (c=='\n')
        {
          yylineno++;
        }
        c=getc(yyin);
      }
      continue;
    }
    /* end of file */
    if (c==EOF)
    {
      return (0);
    } /* end if (c==EOF) */

    /* possible a tag */
    if (c=='<')
    {
      c=getc(yyin);  /* advance to next char */
      if (c==EOF)
      {
        printf("%s(%d) : premature end of file within tag\n",yyerrsrc,yylineno);
        exit(1);
      } /* end if (c==EOF) */

      /* possible a comment */
      if (c=='!')
      {
        /* advance to next char */
        c=getc(yyin);
        if (c!='-')
        {
          /* comment must begin with <!-- */
          printf("%s(%d) : comment format not correct\n",yyerrsrc,yylineno);
          while (c!='>')
          {
            if (c==EOF)
            {
              printf("%s(%d) : premature end of file within tag\n",yyerrsrc,yylineno);
              exit (1);
            } /* end if (c==EOF) */
            if (c=='\n')
            {
              yylineno++;
            }
            c=getc(yyin); /* keep reading and skipping */
          } /* end while (c!='>') */
          c=getc(yyin); /* skip '>' */
          continue;
        } /* end if (c!='-') */
        c=getc(yyin); /* advance to next char */
        if (c!='-')
        {
          /* comment must begin with <!-- */
          printf("%s(%d) : comment format not correct\n",yyerrsrc,yylineno);
          while (c!='>')
          {
            if (c==EOF)
            {
              printf("%s(%d) : premature end of file within tag\n",yyerrsrc,yylineno);
              exit(1);
            } /* end if (c==EOF) */
            if (c=='\n')
            {
              yylineno++;
            }
            c=getc(yyin); /* keep reading and skippintg */
          } /* end while (c!='>') */
          c=getc(yyin); /* skip '>' */
          continue;
        } /* end if (c!='-') */
        /* keep remembering two trailing chars */
        c1=c;
        c2=c1;
        while (c!='>')
        {
          /* scan to the nearest '>' */
          c2=c1;
          c1=c;
          c=getc(yyin);
          if (c==EOF)
          {
            printf("%s(%d) : premature end of file within comment\n",yyerrsrc,yylineno);
            exit(1);
          } /* end if (c==EOF) */
          if (c=='\n')
          {
            yylineno++;
          }
        } /* end while (c!='>') */
        if (c2!='-'||c1!='-')
        {
          /* comment must end with --> */
          printf("%s(%d) : comment format not correct\n",yyerrsrc,yylineno);
        } /* end of (c2!='-'||c1!='-') */
        /* whole comment skipped and continue scanning for next token */
        c=getc(yyin); /* skip '>' */
        continue;
      } /* end if (c=='!') */

      if (isspace(c))
      {
        /* no space allowed between '<' and tag name, otherwise whole tag ignored */
        printf("%s(%d) : tag format not correct, ignore whole tag\n",yyerrsrc,yylineno);
        while (c!='>')
        {
          if (c==EOF)
          {
            printf("%s(%d) : premature end of file within tag\n",yyerrsrc,yylineno);
            exit(1);
          }
          if (c=='\n')
          {
            yylineno++;
          }
          /* whole tag ignored, continue to search next token */
          c=getc(yyin);  /* keep reading and skipping */
        }
        c=getc(yyin); /* skip '>' */
        continue;
      }

      /* this should be a good tag, get it */
      i=0;
      while (c!='>'&&!isspace(c))
      {
        if (c==EOF)
        {
          printf("%s(%d) : premature end of file within tag\n");
          exit(1);
        }
        yytext[i]=c;
        i++;
        c=getc(yyin);
      }
      yytext[i]=0;
      strcpy(yylval.tag.tname,yytext);
      yylval.tag.tparams=0;
      /* because the browser should be case insensitive, 
         all letters convert to uppercase, and search for this tag */
      captalize(yytext);
      t=searchTag(yytext);
      if (t==PRE_BEGIN_TAG)
      {
        isPreformattedText=1;
      }
      if (t==PRE_END_TAG)
      {
        isPreformattedText=0;
      }
      if (t==UNKNOWN_TAG)
      {
        /* this is an illegal tag, just ignore it */
        printf("%s(%d) : unknown tag found, ignore it\n",yyerrsrc,yylineno);
        while (c!='>')
        {
          if (c==EOF)
          {
            printf("%s(%d) : premature end of file within tag\n",yyerrsrc,yylineno);
            exit(1);
          }
          c=getc(yyin);
        }
        c=getc(yyin);
        continue;
      }
      if (c=='>')
      {
        /* this means there is no arguments in tag */
        c=getc(yyin);
        yylval.tag.tparams=0;
        return (t);
      }

      if (isspace(c))
      {
        /* possibly there are arguments in tag */
        while (isspace(c))
        {
          if (c=='\n')
          {
            yylineno++;
          }  
          c=getc(yyin);
        }
        if (c==EOF)
        {
          printf("%s(%d) : premature end of file within tag\n",yyerrsrc,yylineno);
          exit(1);
        }
        if (c=='>')
        {
          /* similarly, space if also not allowed between tag and '>'
             otherwise, it ignored */
          printf("%s(%d) : tag format not correct\n",yyerrsrc,yylineno);
          c=getc(yyin);
          continue;
        }
        /* begin scanning tag arguments */
        s=0;
        while (1)
        {
          i=0;
          while (isalpha(c))
          {
            /* assume that all argument names consists only letters */
            parmtoken[i]=c;
            i++;
            c=getc(yyin);
          }
          parmtoken[i]=0;
          while (isspace(c))
          {
            /* skip possible spaces proceed '=' */
            if (c=='\n')
            {
              yylineno;
            }
            c=getc(yyin);
          }
          if (c!='=')
          {
            /* an argument must be assigned a value by '=' */
            printf("%s(%d) : assign operator missing in parameter\n",yyerrsrc,yylineno);
            while (c!='>')
            {
              if (c==EOF)
              {
                printf("%s(%d) : premature end of file within tag\n",yyerrsrc,yylineno);
                exit(1);
              }
              if (c=='\n')
              {
                yylineno++;
              }
              c=getc(yyin);
            }
            c=getc(yyin);
            goback=1;
            break;
          }
          /* allocate a piece of memory for this argument */
          p=(struct tag_param *)malloc(sizeof(struct tag_param));
          p->pnext=0;
          strcpy(p->pname,parmtoken);
          if (!s)
          {
            /* if this is the first argument, treat it differently */
            s=p;
            q=p;
          }
          else
          {
            q->pnext=p;
            q=p;
          }
          /* now go ahead for the argument value */
          c=getc(yyin);
          /* skip all the possible spaces following '=' */
          while (isspace(c))
          {
            if (c=='\n')
            {
              yylineno++;
            }
            c=getc(yyin);
          }
          if (c==EOF)
          {
            printf("%s(%d) : premature end of file within tag\n",yyerrsrc,yylineno);
            exit(1);
          }
          i=0;
          if (c=='\"')
          {
            /* the case that the value is a string */
            isString=1;
            parmtoken[i]=c;
            i++;
            c=getc(yyin);
          }
          if (isString)
          {
            while (c!='\"')
            {
              switch (c)
              {
                case EOF :
                  printf("%s(%d) : premature end of file within tag.\n",yyerrsrc,yylineno);
                  exit(1);
                default :
                  parmtoken[i]=c;
                  i++;
                  c=getc(yyin);
              }
            }
            isString=0;
            parmtoken[i]=c;
            i++;
            c=getc(yyin);
            yytext[i]=0;
            isString=0;
          }
          else
          {
           /* otherwise the value could be name or a number */
            while (isalnum(c))
            {
              parmtoken[i]=c;
              i++;
              c=getc(yyin);
            }
            parmtoken[i]=0;
            if (c==EOF)
            {
              printf("%s(%d) : premature end of line within tag\n",yyerrsrc,yylineno);
              exit(1);
            }
          }
          strcpy(p->pvalue,parmtoken);
          if (c=='>')
          {
            /* meet the end of tag */
            c=getc(yyin);
            yylval.tag.tparams=s;
            return (t);
          }
          /* otherwise, there are other arguments to be processed */  
          while (isspace(c))
          {
            if (c=='\n')
            {
              yylineno++;
            }
            c=getc(yyin);
          }
          if (c=='>')
          {
            /* end of last argument value should be followed by '>' tightly */
            printf("%s(%d) : tag format not correct\n",yyerrsrc,yylineno);
            c=getc(yyin);
            goback=1;
            break;
          }
          continue;
        }
        if (goback)
        {
          continue;
        } /* end if (goback) */
      }
    }
    i=0;
    /* processing of raw text, whole text between '>' and '<' are treated as TEXT */
    while (c!='<'&&c!=EOF)
    {
      if (c=='&')
      {
        /*following '&' is a character entity */
        c=getc(yyin);
        if (c=='#')
        {
          /* following # should be a decimal */
          c=getc(yyin);
          j=0;
          while (isdigit(c))
          {
            esc[j]=c;
            j++;
            c=getc(yyin);
          }
          esc[j]=0;
          yytext[i]=atoi(esc);
          continue;
        }
        esc[0]='&';   
        j=1;
        while (j<LONGEST_CHAR_ENTITY)
        {
          /* since the character entity names are not in uniform length,
             we need to try all cases within length limit */
          esc[j]=c;
          j++;
          esc[j]=0;
          /* search the character entity table */
          ch=searchChar(esc);
          if (ch==NOT_FOUND)
          {
            /* if could not find, increase the length by one, and continue */
            c=getc(yyin);
            continue;
          }
          else
          {
            c=getc(yyin);
            break;
          }
        }
        if (j==LONGEST_CHAR_ENTITY&&ch==NOT_FOUND)
        {
          for (j=1;j<7;j++,i++)
          {
            yytext[i]=esc[j];
          }
        }
        continue;
      }
      else
      {
        if (c=='\n')
        {
          yylineno++;
          if (isPreformattedText)
          {
            yytext[i]=c;
            i++;
          }
        }
        else
        {
          yytext[i]=c;
          i++;
        }
        c=getc(yyin);
      }
    } /* end while */
    strcpy(yylval.text,yytext);
    return (PLAIN_TEXT);
  } /* end while */
}          
          
/* errprefix - print where and what error message on error file */
static void errprefix(char *msg)
{ 
  int punct=0;

  fprintf(yyerrfile, "[error %d] ", yyerrcnt+1);
  if (yyerrsrc[0]!='\0')                        /* any input file name? */
  {		                
    fprintf(yyerrfile, "file '%s'", yyerrsrc);
    punct=1;
  } /* end if */

  if (yylineno>=0)                         	      /* valid line number? */
  {
    if (punct)
    {
      fprintf(yyerrfile, ", ");
    } /*end if */
    fprintf(yyerrfile, "line %d", yylineno);
    punct=1;
  } /* end if */

  if (yytext==0&&*yytext!='\0') 	             /* real token? */
  {
    if (punct)
    {
      fprintf(yyerrfile, " ");
    } /* end if */
    fprintf(yyerrfile, "near \"%s\"", yytext);
    punct=1;
  } /* end if */

  if (punct)
  {
    fprintf(yyerrfile, ": ");
  } /* end if */
  fprintf(yyerrfile, "%s\n", msg);
} /* end errprefix */

/* yyerror - improved error reporting */
void yyerror(char *s, char *t)
{ 
  static const char *expecting="expecting: ";
  static int list=0;
  static int column=0;

  if (s!=0) 
  {
    if (column!=0)
    {
      fputc('\n', yyerrfile);
    } /*end if */
    errprefix(s);
    if (t==0)
    {
      column=0;
    }
    else
    {
      column=fprintf(yyerrfile,"actual: %s", t);
    } /* end if */
    list=0;
  } 
  else
  {
    if (t!=0) 
    {
      if (list == 0) 
      {
        if (column+strlen(t)+sizeof(expecting)+1<LINEWIDTH-2)
        {
          column+=fprintf(yyerrfile, "  %s%s", expecting, t);
        }
        else
        {
          column=fprintf(yyerrfile, "\n%s%s", expecting, t)-1;
    	} /* end if */
      }
      else 
      {
        if (column+strlen(t)<LINEWIDTH-2)
        {
          column+=fprintf(yyerrfile, ", %s", t);
        }
        else
        {
          column=fprintf(yyerrfile, ",\n    %s", t)-1;
        } /* end if */
      } /* end if */
      ++list;
    }
    else 
    {
      fprintf(yyerrfile,"\n");
      column=0;
      list=0;
    } /* end if */
  } /* end if */
} /* end yyerror */
                          

char* yydisplay(int tokenValue)
{
  int i;
  char buf[20];

  if (tokenValue>YYERRCODE)
  {
    i=0;
    while (tagList[i].v!=tokenValue)
    {
      i++;
    }
    return (tagList[i].n);
  }
  
  switch (tokenValue)
  {
    case '\a': return ("[\\a]");
    case '\b': return ("[\\b]");
    case '\v': return ("[\\v]");
    case '\f': return ("[\\f]");
    case '\r': return ("[\\r]");
    case '\t': return ("[\\t]");
    case '\n': return ("[\\n]");
  }

  if (tokenValue==0)
  {
    return ("[end of file]");
  }
  if (tokenValue==YYERRCODE)
  {
    return ("[error]");
  }

  if (isprint(tokenValue))
  {
    sprintf(buf,"%c",tokenValue);
    return (buf);
  }
  else
  {
    sprintf(buf,"[char %d]",tokenValue);
    return (buf);
  }
}
 
